# -*- coding: utf-8 -*-


class SKTBaseException(Exception):
    """Base Exception for vespene"""


class AppConfigLoadFailException(SKTBaseException):
    """Failed to load app.yaml, not found or yaml invalid"""


class ParameterErrorException(SKTBaseException):
    """Bad parameter"""


class EnvNotReadyYetException(SKTBaseException):
    """Env is not ready yet"""


class WSGIException(SKTBaseException):
    """WSGI related Exception"""



# -*- coding: utf-8 -*-

import logging

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter(
    logging.Formatter('[SKT %(levelname)-7s] %(message)s')
)
logger.addHandler(console)
logger.setLevel(logging.INFO)



# -*- coding: utf-8 -*-

from redis import Redis


def create_redis_client(redis_dsn):
    '''Create redis client from dsn uri
    :param redis_dsn: Redis uri, such as "redis://localhost:6383"
    :return: `Redis` instance
    '''
    return Redis.from_url(redis_dsn)


def create_redis_client_from_config(config):
    '''Create redis client from config dict
    :param config: Redis confis, such as {"host": "localhost", "port": 6383}
    :return: `Redis` instance
    '''
    return Redis(**config)



# -*- coding: utf-8 -*-

APP_CONFIG_PATH = "app.yaml"
ARCH_CONFIG_PATH = "/etc/rest_arch/env.yaml"

APP_TYPE_BARE = "bare"

HUSKAR_CACHE_DIR = "/var/run/skt_huskar"

DEFAULT_APP_PORT = 8010

DEFAULT_WORKER_NUMBER_DEV = 1
DEFAULT_WORKER_NUMBER_PROD = 2
DEFAULT_WORKER_CLASS = 'gevent'

ENV_DEV = 'dev'
ENV_TESTING = 'testing'
ENV_PROD = 'prod'



# -*- coding: utf-8 -*-

import sys
import yaml
import logging

from .consts import (
    ARCH_CONFIG_PATH,
    APP_CONFIG_PATH,
    APP_TYPE_BARE,
    HUSKAR_CACHE_DIR,
    DEFAULT_APP_PORT,
    DEFAULT_WORKER_CLASS,
    DEFAULT_WORKER_NUMBER_DEV,
    DEFAULT_WORKER_NUMBER_PROD,
    ENV_DEV
)
import rest_arch
from .exc import AppConfigLoadFailException
from ..utils import EnvvarReader, cached_property


logger = logging.getLogger(__name__)


class ArchConfig(object):
    """
    Application independent configs, as a yaml file at
    `CORE_CONFIG_PATH`, example::

        env: 'dev'
        cluster: 'arch-sync-01'
        zookeeper_config:
          hosts: localhost:2181
          username: root
          password:

    """

    def __init__(self):
        self.config = None

    def _load(self, _file):
        try:
            if isinstance(_file, basestring):
                self.config = yaml.load(open(_file))
            else:
                self.config = yaml.load(_file)
        except (IOError, yaml.error.YAMLError):
            logger.warn("cannot load config from %r, useing default", _file)
            self.config = {}

    def load(self, file=None):
        file = file or ARCH_CONFIG_PATH
        self._load(file)
        return self

    @property
    def env(self):
        return self.config.get('env', ENV_DEV)

    @property
    def cluster(self):
        return self.config.get('cluster', None)

arch_config = None


def load_arch_config():
    global arch_config
    if arch_config is None:
        arch_config = ArchConfig().load()
    return arch_config

haproxy_port_map = {}
worker_num_map = {}
port_map = {}


class BareAppConfig(object):
    """Application related configs, as a yaml file at
    `APP_CONFIG_PATH`, e.g.:

        app_name: ves.note
        settings: note.settings
        services:
          wsgi:
            app: note:app
            worker_class: gevent
            requirements: wsgi_requirements.txt
    """

    TYPE_NAME = APP_TYPE_BARE
    PORT_MAP = port_map
    DEFAULT_APP_PORT = None
    DEFAULT_WORKER_CLASS = None

    envvar = EnvvarReader(
        "SKT_APP_PORT",
        "SKT_APP_HAPROXY_PORT",
        "SKT_APP_WORKER_NUM"
    )

    def __init__(self):
        self.config = None

    def load(self, config_path=APP_CONFIG_PATH, raise_exc=False):
        try:
            self.config = yaml.load(open(config_path))
        except (IOError, yaml.error.YAMLError):
            if raise_exc is True:
                raise AppConfigLoadFailException
            logger.error('cannot load %s, exit', config_path)
            sys.exit(1)
        return self

    @cached_property
    def app_name(self):
        return self.config["app_name"]

    @cached_property
    def app_id(self):
        return self.app_name

    @cached_property
    def app_slug(self):
        return self.app_name.rsplit('.', 1)[-1]

    @cached_property
    def app_uri(self):
        app = self._get_conf('app', None)
        if app is None:
            raise RuntimeError("Missing `app` in app.yaml")
        return app

    @cached_property
    def app_settings_uri(self):
        return self.config['settings']

    @cached_property
    def port(self):
        from .env import is_in_dev

        port = self.envvar.get_int('skt_app_port')
        if port:
            return port
        if is_in_dev():
            return self.DEFAULT_APP_PORT
        return self.PORT_MAP.get(self.app_name, self.DEFAULT_APP_PORT)

    @cached_property
    def canonical_port(self):
        from .env import env, is_in_prod

        if not is_in_prod():
            port = self.port
        else:
            port = self.envvar.get_int('skt_app_haproxy_port')
            if not port:
                port = haproxy_port_map.get(self.app_name)
        if not port:
            raise RuntimeError(
                'no canonical port can be determined for app: %r in env: %r'
                % (self.app_name, env()))
        return port

    def get_app_n_workers(self):
        from .env import is_in_prod

        worker_num = self.envvar.get_int('skt_app_worker_num')
        if worker_num:
            return worker_num
        if not is_in_prod():
            return DEFAULT_WORKER_NUMBER_DEV
        return worker_num_map.get(self.app_name, DEFAULT_WORKER_NUMBER_PROD)

    @cached_property
    def worker_class(self):
        return self._get_conf('worker_class', self.DEFAULT_WORKER_CLASS)

    def get_app_binds(self):
        return '0.0.0.0:%d' % self.port

    @cached_property
    def logger_name(self):
        return self.app_name.split('.')[-1]

    @cached_property
    def arch_deps(self):
        with open(self.requirements_file) as f:
            arch_version = None
            for line in f:
                if "rest_arch" in line:
                    arch_version = line.strip()
        return arch_version or "rest_arch==%s" % rest_arch.__version__

    @cached_property
    def message_consumers(self):
        consumers = self.config.get('message_consumer')
        if not consumers:
            raise AppConfigLoadFailException(
                'message.consumer should contains something in app.yaml.')

        return consumers

    @cached_property
    def message_consumers_settings(self):
        return self.config.get('message_consumer_settings', {})

    @cached_property
    def huskar_app_name(self):
        return self.app_name

    @cached_property
    def async_workers(self):
        return self.config.get('async_workers')

    @cached_property
    def requirements_file(self):
        return self.config.get('requirements', 'requirements.txt')

    @cached_property
    def timeout(self):
        return self._get_conf('timeout')

    def get_service_register(self):
        return 'huskar_sdk:ServiceWatcher'

    def get_service_register_conf(self):
        load_arch_config()
        conf = {}
        conf.update(arch_config.zookeeper_config)
        conf.update({
            'team': 'arch',
            'cluster': arch_config.cluster,
            'service_name': self.app_name,
            'cache_dir': HUSKAR_CACHE_DIR,
        })
        return conf

    def _get_conf(self, key, default=None):
        """Help to try config frist on key 'services' and then root.
        """
        if 'services' in self.config and \
                self.TYPE_NAME in self.config['services']:
            return self.config['services'][self.TYPE_NAME].get(key, default)
        return self.config.get(key, default)


class WsgiAppConfig(BareAppConfig):
    TYPE_NAME = "wsgi"
    DEFAULT_APP_PORT = DEFAULT_APP_PORT
    DEFAULT_WORKER_CLASS = DEFAULT_WORKER_CLASS


app_config = None


def load_app_config(raise_exc=False):
    global app_config
    if app_config is None:
        app_config = WsgiAppConfig().load(raise_exc=raise_exc)
    return app_config



# -*- coding: utf-8 -*-

from ..env import is_in_dev
from ..config import load_app_config, load_arch_config
from gunicorn.app.wsgiapp import WSGIApplication
from . import hooks


class SetAppMixin(object):
    def arch_set(self, config):
        self.cfg.set('default_proc_name', config.app_name)
        self.cfg.set('worker_class', config.worker_class)
        self.cfg.set('loglevel', 'info')
        self.cfg.set('graceful_timeout', 3)
        self.cfg.set('bind', config.get_app_binds())
        self.cfg.set('workers', config.get_app_n_workers())

        if is_in_dev():
            self.cfg.set('errorlog', '-')
        else:
            self.cfg.set('syslog', True)
            self.cfg.set('syslog_facility', 'local6')
            self.cfg.set('syslog_addr', 'unix:///dev/log#dgram')


class ArchWsgiApp(SetAppMixin, WSGIApplication):

    def init(self, parser, opts, args):
        self.app_config = load_app_config()
        self.app_uri = self.app_config.app_uri
        self.app_settings_uri = self.app_config.app_settings_uri
        self.arch_config = load_arch_config()
        args = [self.app_uri]
        self.arch_set(self.app_config)
        self.cfg.set('syslog_prefix', "{0}.wsgi".format(
            self.app_config.app_name))
        self.cfg.set('post_fork', hooks.post_fork)
        self.cfg.set('accesslog', '-')
        self.cfg.set('errorlog', '-')
        self.cfg.set('statsd_prefix',
                     "skt.wsgi.%s" % self.app_config.app_name)
        super(ArchWsgiApp, self).init(parser, opts, args)


def serve():
    ArchWsgiApp().run()



# -*- coding: utf-8 -*-

import functools
import logging
import os


class Service(dict):

    def __setattr__(self, name, value):
        return super(Service, self).__setitem__(name, value)

    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)

    def dispather(self, cls):
        def _decorate(func):
            @functools.wraps(func)
            def wrapper(cls_self, *args):
                return self.run(cls_self, func, args, logger)
            return wrapper

        logger = logging.getLogger(cls.__module__)

    def get_dispatcher_class(self):
        if self.thrift_service_dispatcher_cls is None:
            raise RuntimeError("Dispatcher is still not registered")
        return self.thrift_service_dispatcher_cls

    def _is_shutting_down(self):
        """
        If gunicorn wants worker to gracefully shutdown, it
        will call `worker_term`(in gunicorn_config.py). We set
        a system env there, and at this place, check that env.
        """
        return os.environ.get('SERVER_SHUTTINGDOWN')

    def _raise_unknown_error(self, exc):
        """
        Help to raise an `unknown_exc` of current service.
        """
        code = self.error_code.UNKNOWN_ERROR
        name = self.error_code._VALUES_TO_NAMES[code]
        raise self.unknown_exc(code, name, repr(exc))

    def run(self, dispatcher, func, args, logger):
        pass



# -*- coding: utf-8 -*-

import logging
import os
import sys
from .config import load_arch_config, load_app_config
from .exc import AppConfigLoadFailException, EnvNotReadyYetException
from .consts import (
    APP_CONFIG_PATH,
    ENV_DEV,
    ENV_TESTING,
    ENV_PROD
)

logger = logging.getLogger(__name__)


def env():
    return load_arch_config().env


def is_in_dev():
    return load_arch_config().env == ENV_DEV


def is_in_testing():
    return load_arch_config().env == ENV_TESTING


def is_in_prod():
    return load_arch_config().env == ENV_PROD


_is_in_container = None


def is_in_container():
    global _is_in_container
    if _is_in_container is None:
        if os.path.exists(APP_CONFIG_PATH):
            try:
                load_app_config(raise_exc=True)
            except AppConfigLoadFailException:
                _is_in_container = False
            else:
                _is_in_container = True
        else:
            _is_in_container = False
    return _is_in_container

gevent_patched = False
in_profile = False
loggers_initialized = False
initialized = False


def set_profile(v):
    global in_profile
    in_profile = bool(v)


def is_in_profile():
    return in_profile


def require(name):
    if is_in_container() and not globals()[name]:
        raise EnvNotReadyYetException(
            "env not ready: require {}".format(name))


def initialize():
    """Initialize worker process environment, including: gevent patch,
    settings, signals, logging, client pools etc. It should be called
    before all other operations are taken, right after a worker is forked
    or in cmd entry points. Please note the order of these operations."""
    global initialized
    if initialized:
        logger.warn("Env is already initialized, skipping")
        return
    patch_gevent()
    initialized = True


def patch_gevent():
    """Call gevent monkey patch right after ``post_fork``.
    """
    global gevent_patched
    if gevent_patched:
        logger.warn("Gevent is already patched, skipping")
        return
    if 'threading' in sys.modules:
        logger.warn('Threading imported before gevent patching.')
    import gevent.monkey
    gevent.monkey.patch_all()
    gevent_patched = True



# -*- coding: utf-8 -*-


class ServerRestartException(Exception):
    """Raised if a `ping` request incomes but our sever is shutting down."""
    pass


class ArchWarningException(Exception):
    pass


###
# Message's Exc
###

class ArchMessageExc(Exception):
    """ Message base exc """


class BrokersInfoNotFoundExc(ArchMessageExc):
    message = 'Could not found mq\' brokers info in app.yaml'


class ConsumerConfigBadFormatExc(ArchMessageExc):
    message = 'Message consumer\'s config is not correctly formatted ' \
              'in app.yaml.'


class BrokerUrlsNotConfigedExc(ArchMessageExc):
    message = 'Message queue\'s broker urls is not configured in app.yaml'



# -*- coding: utf-8 -*-

import sys
import contextlib
import click


@contextlib.contextmanager
def _client_ctx(app_name, service, host, port):
    pass


@click.command()
@click.option('-h', '--host', default='localhost',
              help='service host(default localhost)')
@click.option('-p', '--port', default=None, help='service port')
@click.option('-r', '--rpc', default=False, is_flag=True,
              help='if use rpc connect')
@click.option('--profile', default=False, is_flag=True,
              help='Profile all apis calls')
def shell(host, port, rpc, profile):
    def embed_with_cli():
        import IPython
        del sys.exitfunc
        IPython.embed()
    embed_with_cli()



# -*- coding: utf-8 -*-

"""
Hooks around gunicorn runner.
"""

import os


def worker_term(worker):
    os.environ['SERVER_SHUTTINGDOWN'] = '1'


def post_worker_init(worker):
    os.environ.pop('SERVER_SHUTTINGDOWN', None)


def on_connected(worker, addr):
    from ...ctx import g
    if addr:
        g.set_conn_meta('client_addr', addr[0])


def post_fork(arbitor, worker):
    worker.app.chdir()
    from ..env import initialize
    initialize()


def when_ready(arbitor):
    pass



# -*- coding: utf-8 -*-

import re
import os
import time
import datetime
import hashlib
import uuid

# used for mobile verify
MOBILE_RE = re.compile('^1[34578]\d{9}$')

ALL_PHONE_RE = re.compile('^1[34578]\d{9}$|^[2-9]\d{6,7}(-\d{1,4})?$'
                          '|^6[1-9]{2,5}$')

# used for email verify
EMAIL_RE = re.compile('\w+([-+.]\w+)*@\w+([-.]\w+)*\.\w+([-.]\w+)*')

# used for number translate
UNIT_MAP = {u'十': 10, u'百': 100, u'千': 1000, u'万': 10000, u'亿': 100000000}
NUM_MAP = u"零一二三四五六七八九"

# used for capturing timeout error
TServerTimeoutSentryProject = None

# /path/to/rest_arch/rest_arch
LIB_DIR_PATH = os.path.dirname(os.path.abspath(__file__))


def is_mobile(s):
    """
    Check whether a string is mobile number.

    :param s: mobile string to be checked
    """
    return MOBILE_RE.match(s) is not None


def is_phone(s):
    """
    Check whether a string is a fixed phone format.

    :param s: phone string to be checked
    """

    return ALL_PHONE_RE.match(s) is not None


def is_email(s):
    """
    Check whether a string is email address.

    :param s: email string to be checked
    """
    return EMAIL_RE.match(s) is not None


def datetime2utc(dt):
    """
    Sends a :class:`datetime.datetime` object.
    Returns :class:`int` time value.

    :param dt: :class:`datetime` object
    """
    return int(time.mktime(dt.timetuple()))


def utc2datetime(t):
    """
    Sends a :class:`float` time value.
    Returns :class:`datetime.datetime` object.

    :param t: :class:`float` time value.
    """
    return datetime.datetime.fromtimestamp(t)


def strptime2date(date_str):
    return datetime.datetime.strptime(date_str, '%Y-%m-%d').date()


def chinese2num(s):
    n = value = unit = flag = 0
    for i in range(len(s)):
        if s[i] in NUM_MAP:
            value = NUM_MAP.index(s[i])
            n += value * unit / 10 if flag and i == len(s) - 1 else value
            flag = 0
        elif s[i] in UNIT_MAP:
            unit = flag = UNIT_MAP[s[i]]
            if unit >= 10000:
                return n * unit + chinese2num(s[i + 1:])
            n += (unit - 1) * value if value else unit
        else:
            raise ValueError
    return n


def md5(s):
    """
    Returns :class:`str` object after md5 encrypt.

    :param s: :class:`str`, the str to md5 encrypt.
    """
    return hashlib.new('md5', s).hexdigest()


def sha1(s):
    """
    Returns :class:`str` object after sha1 encrypt.

    :param s: :class:`str`, the str to sha1 encrypt.
    """
    return hashlib.new('sha1', s).hexdigest()


def get_unique_string(length=31):
    uid = uuid.uuid1()
    full = md5(uid.hex)
    if length > 0:
        return full[:length]
    else:
        return full


class CachedProperty(object):
    """Decorator like python built-in ``property``, but it results only
    once call, set the result into decorated instance's ``__dict__`` as
    a static property.
    """
    def __init__(self, fn):
        self.fn = fn
        self.__doc__ = fn.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        val = inst.__dict__[self.fn.__name__] = self.fn(inst)
        return val

# alias
cached_property = CachedProperty


class EnvvarReader(dict):
    __uninitialized = object()

    def __init__(self, *envvars):
        for ev in envvars:
            ev_norm = self._normalize(ev)
            if ev != ev_norm:
                raise ValueError(
                    'envvar name should be UPPERCASED: {!r}'.format(ev))
            dict.__setitem__(self, ev_norm, self.__uninitialized)

    @cached_property
    def configured(self):
        return any([self[k] for k in self])

    def _normalize(self, varname):
        return varname.strip().upper()

    def __getitem__(self, key):
        key = self._normalize(key)
        value = dict.__getitem__(self, key)
        if value is self.__uninitialized:
            value = os.getenv(key, None)
            dict.__setitem__(self, key, value)
        return value

    __getattr__ = __getitem__

    def get_int(self, key):
        value = self.__getitem__(key)
        if value is not None:
            return int(value)

    def __setitem__(self, key, value):
        raise RuntimeError('setting envvar not supported')

    __setattr__ = __setitem__



# -*- coding: utf-8 -*-

import logging
import click

from .serve import serve
from .shell import shell


logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
def skt():
    pass

skt.add_command(serve)
skt.add_command(shell)



# -*- coding: utf-8 -*-

from flask import Flask


##
# make flask app with param config
##
def make_app(name, **config):
    global _app_activated
    try:
        _app_activated = True
        return Flask(name, **config)
    except Exception as exc:
        raise SystemExit(exc)



# -*- coding: utf-8 -*-

import mock
import rest_arch.app as app


def test_app():
    mock_obj = mock.Mock(return_value="app")
    with mock.patch.object(app, 'make_app', mock_obj):
        res = app.make_app("test")
        assert res == "app"



# -*- coding: utf-8 -*-

import os
import re
from setuptools import setup, find_packages


def _get_version():
    v_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               'rest_arch', '__init__.py')
    ver_info_str = re.compile(r".*version_info = \((.*?)\)", re.S). \
        match(open(v_file_path).read()).group(1)
    return re.sub(r'(\'|"|\s+)', '', ver_info_str).replace(',', '.')

entry_points = [
    # skt
    "skt = rest_arch.skt.cmds:skt"
]

data = [
    "static/*",
    "skt/static/*"
]

setup(
    name="rest_arch",
    version=_get_version(),
    description="A Restful Framework Based On Flask",
    long_description=open("README.md").read(),
    author="kiven",
    author_email="kiven.mr@gmail.com",
    packages=find_packages(),
    package_data={"": ["LICENSE"], "skt": data},
    url="https://github.com/MrKiven/REST_ARCH/",
    tests_require=[
        'pytest==2.5.2',
        'pytest-cov==1.8.1',
        'pytest-xdist==1.13.1',
        'mock==1.0.1',
    ],
    entry_points={"console_scripts": entry_points},
    install_requires=[
        'PyMySQL==0.6.2',
        'redis==2.10.3',
        'requests==2.7.0',
        'SQLAlchemy==0.9.3',
        'PyYAML==3.11',
        'ipython==3.1.0',
        'gipc==0.5.0',
        'wheel==0.24.0',
        'click==5.1',
        'Fabric==1.10.2',
        'Jinja2==2.7.3',
    ],

)



# -*- coding: utf-8 -*-

import click


@click.command(
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True
    },
    add_help_option=False
)
@click.pass_context
def serve(ctx):
    from .. import runner
    runner.serve()


if __name__ == '__main__':
    # pylint: disable=E1120
    serve()
